/*
     Program: COLORPAL()
     System: GRUMPFISH LIBRARY
     Author: Greg Lief
     Copyright (c) 1988-93, Greg Lief
     CA-Clipper 5.x Version
     Compile instructions: clipper colorpal /n /w
     *Mouseable*
     Allows user to select color from interactive palette
     Procs & Fncts: COLORPAL()
     Calls: COLOR_N2S() * COLOR_S2N()  (see COLORS.PRG)
*/

#include "inkey.ch"
#include "grump.ch"

#define INDICATOR 10  // ASCII char to use as indicator when moving thru palette
#define NUM_COLORS  256

function colorpal(cCurrentColor, nTop, nLeft, bRedraw)
static cPalette
local nRow
local nCol
local lEscape := .f.
local lContinue := .t.
local xx
local lUseSample := .f.
local lFirstLoop := .t.
local cBuffer
local nBottom
local nKey

#ifdef MOUSE
local nMRow
local nMCol
#endif

GFSaveEnv(.t., 0)           // shut off cursor

// establish top row and leftmost column for palette if not passed
if nTop == NIL .or. nTop > maxrow() - 16
   nTop := maxrow() - 16
endif
nBottom  := nTop + 16
if nLeft == NIL .or. nLeft > maxcol() - 15
   nLeft := maxcol() - 15
endif

/*
   establish color sample UDF.  there are three options:
   a) user did not pass fourth parameter --- use default
   b) user passed a code block as fourth parameter - use that unchanged
   c) user passed character string as parameter (such as "whatever()") --
      compile it to a code block
*/
if valtype(bRedraw) == 'C'
   bRedraw := MakeBlock(bRedraw)
elseif valtype(bRedraw) <> 'B'
   lUseSample := .T.
   bRedraw := { | a, b | colorsample(a, b) }
endif

// set start-up color to current color if no parameter passed
cCurrentColor := if(cCurrentColor == NIL, setcolor(), upper(cCurrentColor))

// convert the string to a number
cCurrentColor := color_s2n(cCurrentColor)

// create character string of text/color attributes for color palette
// if it was not already created on a previous visit to COLORPAL()
if cPalette == NIL
   cPalette := []
   for xx := 0 to NUM_COLORS
      cPalette += chr(4) + chr(xx)
   next
endif
restscreen(nTop, nLeft, nBottom - 1, nLeft+15, cPalette)
ColorSet(C_MESSAGE)
scroll(nBottom, nLeft, nBottom, nLeft+15, 0)
@ nBottom, nLeft      ssay chr(24)+chr(25)+chr(27)+chr(26)
@ nBottom, nLeft + 7  ssay chr(17) + chr(217)
@ nBottom, nLeft + 13 ssay 'Esc'

// determine starting row and column within palette
nRow := nTop + int(cCurrentColor / 16)
nCol := nLeft + cCurrentColor % 16

// commence main keypress loop
do while lContinue .and. ! lEscape

   // change to currently selected color, but if it is a blinking
   // color, use the non-blinking component (will reset below)
   if cCurrentColor > 127 .and. setblink()
      setcolor(Color_N2S(cCurrentColor - 128))
   else
      setcolor(Color_N2S(cCurrentColor))
   endif

   // make sure color sample gets drawn in proper color first time thru
   if lFirstLoop .and. lUseSample
      colorsample( nTop, nLeft )
      lFirstLoop := .f.
   endif

   // draw indicator to mark current color
   @ nRow,nCol ssay chr(INDICATOR)

   // if currently selected color is blinking, must reset it (see above)
   if cCurrentColor > 127 .and. setblink()
      setcolor(Color_N2S(cCurrentColor))
   endif

   nKey := ginkey( 0 )

   // clear indicator
   @ nRow,nCol ssay chr(4)

   // process keystroke / mouse click
   do case
      #ifdef MOUSE
      case nKey == K_LEFT_DOWN
         nMRow := GFLastRow()
         nMCol := GFLastCol()

         if nMRow >= nTop .and. nMRow <= nTop + 15 .and. nMCol >= nLeft .and. nMCol <= nLeft + 15
            if nRow <> nMRow
               cCurrentColor += ( ( nMRow - nRow ) * 16 )
               nRow := nMRow
            endif

            if nCol <> nMCol .and. nMCol >= nLeft .and. nMCol <= nLeft + 15
               cCurrentColor += ( nMCol - nCol )
               nCol := nMCol
            endif
         endif

      case nKey == K_LEFT_DBLCLICK
         nMRow := GFLastRow()
         nMCol := GFLastCol()
         if nMRow >= nTop .and. nMRow <= nTop + 15 .and. nMCol >= nLeft .and. nMCol <= nLeft + 15
            lContinue := .F.
         endif

      case nKey == K_RIGHT_DOWN
         lEscape := .t.
      #endif

      case nKey == K_DOWN
            // if we are at the bottom, jump to the top
         IF nRow == nBottom - 1
            cCurrentColor -= (nBottom - 1 - (nRow := nTop) ) * 16
         else
            nRow++
            cCurrentColor += 16
         endif

      case nKey == K_UP
         // if we are at the top, jump to the bottom
         if nRow == nTop
            cCurrentColor += ( (nRow := nBottom - 1) - nTop) * 16
         else
            nRow--
            cCurrentColor -= 16
         endif

      case nKey == K_RIGHT
         if nCol < nLeft + 15
            nCol++
            cCurrentColor++
         else
            nCol := nLeft
            cCurrentColor -= 15
         endif

      case nKey == K_LEFT
         if nCol > nLeft
            nCol--
            cCurrentColor--
         else
            nCol := nLeft + 15
            cCurrentColor += 15
         endif

      case nKey == K_PGDN
         cCurrentColor += (nBottom - 1 - nRow) * 16
         nRow := nBottom - 1

      case nKey == K_PGUP
         cCurrentColor -= (nRow - nTop) * 16
         nRow := nTop

      case nKey == K_ENTER
         lContinue := .f.

      case nKey == K_ESC
         lEscape := .t.

   endcase

   if lContinue .and. ! lEscape
      setcolor(color_n2s(cCurrentColor))
      dispbegin()
      if ! lUseSample
         cBuffer := savescreen(nTop, nLeft, nBottom, nLeft+15)
         #ifdef MOUSE
         HideMouseCursor()
         #endif
      endif
      eval(bRedraw, nTop, nLeft)
      if ! lUseSample
         restscreen(nTop, nLeft, nBottom, nLeft+15, cBuffer)
         #ifdef MOUSE
         ShowMouseCursor()
         #endif
      endif
      dispend()
   endif
enddo

// restore environment
GFRestEnv()

// if user pressed ESC or right mouse button, return the old color
return (if(lEscape, setcolor(), color_n2s(cCurrentColor)))

* end function ColorPal()
*--------------------------------------------------------------------*


/*
   ColorSample(): draw sample box showing current selection
*/
static function ColorSample(nTop, nLeft)
DOUBLEBOX(nTop-6, nLeft, nTop-3, nLeft+15)
@ nTop-5, nLeft+3 ssay "Sample of"
@ nTop-4, nLeft+1 ssay "current color"
return NIL

* end static function ColorSample()
*--------------------------------------------------------------------*

* eof colorpal.prg

