/*
   Function:   GFPICKFILE()
   Author:     Greg Lief
   Copyright:  (c) 1993 Greg Lief
   Dialect:    CA-Clipper 5.x
   Purpose:    Primary mechanism for selecting databases/indexes
   Syntax:     GFPICKFILE( [<cExt>] [, <lShowstru>] [, <lRebuild>] )
   Parameters: <cExt> - file extension to pick from (default: "DBF")

               <lShowstru> - numeric indicating which key will allow the
               user to view the structure of the highlighted DBF or NTX.
               If not passed, this feature will not be available.  Note
               that this will only work if you pass it the "DBF" or
               current index file extension.

               <lRebuild> - pass .F. if you do not want the files array
               to be rebuilt.  (By default it will be rescanned every
               time you call this function.)

   Returns:        Complete path and filename of selected file
   Support Funcs:  ChangeDrive() - allow user to change drive
                   ShowStru()    - show structure of highlighted DBF
                   ShowKey()     - show index key of highlighted file

   Warning:    As written, this function will not pick up any directory
               entries that have extensions (e.g., JUNK.DIR).  If you
               have such directories, change the "*." in line 83 below
               to "*.*" so that it reads:

               aFiles := directory(cPath + "*.*", "D")
*/

#include "apick.ch"
#include "dbstruct.ch"
#include "directry.ch"
#include "fileio.ch"
#include "grump.ch"
#include "inkey.ch"

// manifest constants to cut down on clutter
#define FILENAME   aFiles[x][F_NAME]
#define FILEATTR   aFiles[x][F_ATTR]

function GFPickFile(cExt, nStruKey, lRebuild)
local aTemp
local x
local y
local nFiles
local bOldF10
local bOldStruKey
local cRetval
local lDriveChanged
local lShowMsg
local cFileSpec := "*." + if(cExt == NIL, "DBF", cExt)
static aFiles                 // array of files for selection
static cPath                  // current path
static cLastExt               // last extension used (.DBF or .NTX)...
                              // If you call PICKFILE() with a different
                              // extension than the previous time, we
                              // must rebuild the files array (see below)
if cPath == NIL
   cPath := gfdrive() + ":\" + if(! empty(curdir()), curdir() + "\", '')
endif
default cExt to "DBF"
default lRebuild to .t.
if ! lRebuild
   lRebuild := (cFileSpec <> cLastExt)
endif
bOldF10 := setkey(K_F10, { | | lDriveChanged := changedrive(@cPath) } )
if nStruKey <> NIL
   if cExt == substr(indexext(), 2)
      bOldStruKey := setkey(nStruKey, ;
                     { | p, l, ele | showkey(cPath + aFiles[val(ele)]) } )
   elseif cExt == "DBF"
      bOldStruKey := setkey(nStruKey, ;
                     { | p, l, ele | showstru(cPath + aFiles[val(ele)]) } )
   endif
endif
do while cRetval == NIL
   if lRebuild .or. aFiles == NIL

      // first, must grab all files to ensure that we get all directories
      aFiles      := directory(cPath + "*.", "D")
      nFiles      := len(aFiles)
      lShowMsg    := (nFiles > 30)

      // show status message if there are more than a few files
      if lShowMsg
         waiton( { "Scanning files..." }, .f.)
      endif

      // strip out all non-directory entries
      for x := nFiles to 1 step -1
         // non-directory entry
         if FILEATTR <> "D" .or. FILENAME == "."
            adel(aFiles, x)
            nFiles--
         // manipulate filenames for directory entries
         elseif FILENAME <> ".."
            FILENAME := "<" + FILENAME + ">"
         endif
      next

      // then grab desired files based on global filespec
      aTemp := directory(cPath + cFileSpec)

      // resize aFiles array to required length
      y := nFiles + len(aTemp)
      asize(aFiles, y)

      // copy file subarrays into master array
      for x := nFiles + 1 to y
         aFiles[x] := aTemp[x - nFiles]
      next

      // convert aFiles into single-D array, keeping only filename
      for x := 1 to y
         aFiles[x] := FILENAME
      next
      asort(aFiles)
      if lShowMsg
         waitoff()
      endif
   else
      nFiles := len(aFiles)
   endif
   lRebuild := .t.                // leave this on for subsequent loops
   // if no files or subdirectories, force user to change drive
   if empty(aFiles)
      changedrive(@cPath, .t.)
      #ifdef MOUSE
      if lastkey() == K_ESC .or. lastkey() == K_RIGHT_DOWN
      #else
      if lastkey() == K_ESC
      #endif
         cRetval := ""            // force exit from main loop
      endif
   else
      // the lDriveChanged flag is necessary because ChangeDrive()
      // stuffs the keyboard with ESC to short-circuit APICK().
      // We must therefore be able to tell the difference between
      // a bogus escape and a real escape.
      lDriveChanged := .f.
      // restrict APICK() box size so it doesn't go over msg box
      if nFiles > maxrow() - 11
         apick x array aFiles top 4 bottom maxrow() - 6 title cPath
      else
         apick x array aFiles title cPath
      endif
      if ! lDriveChanged
         #ifdef MOUSE
         if lastkey() <> K_ESC .and. lastkey() <> K_RIGHT_DOWN
         #else
         if lastkey() <> K_ESC
         #endif
            do case
               // parent directory
               case aFiles[x] == ".."
                  // step backwards by two backslashes!
                  cPath := substr(cPath, 1, rat("\", cPath) - 1)
                  cPath := substr(cPath, 1, rat("\", cPath))
               // subdirectory
               case "<" $ aFiles[x]
                  cPath += substr(aFiles[x], 2, len(aFiles[x]) - 2) + '\'
               // a file!
               otherwise
                  cRetval := cPath + aFiles[x]
            endcase
         else
            cRetval := ""         // force exit from main loop
         endif
      endif
   endif
enddo
setkey(K_F10, bOldF10)
if nStruKey <> NIL
   setkey(nStruKey, bOldStruKey)
endif
cLastExt := cFileSpec   // save this for next time... see notes above
return cRetval

* end function PickFile()
*--------------------------------------------------------------------*


/*
   ChangeDrive( <cPath> [, <lMustchange> ] )

   Allow user to search for files on different drive.
   Special thanks to Joe Booth for the NUL idea!

   <cPath> is passed by reference so that it can be changed directly by
   this function.

   <lMustchange> is a flag indicating that the user was forced to change
   drives, probably due to lack of DBF files on the current drive.

   Returns true (.T.) if drive was actually changed, otherwise false (.F.)
*/
static function ChangeDrive(cPath, lMustChange)
local cDrive
local nHandle
local nEle
local lRetval := .f.
local lChoice := .t.
static cLastPath   // see below for notes about this variable

// the CURDIR_ array holds current directory for each drive so that
// when you bounce between drives, you won't start over at the root
static aCurDir := {}

// it is possible to change to a drive that has no DBFs... cLastPath
// keeps track of the previous path so that we can automatically
// go back to it rather than force the user to enter another drive
if lMustChange <> NIL
   err_msg({ "No files found on this drive " })
   if cLastPath <> NIL .and. cLastPath <> cPath
      cPath     := cLastPath
      cLastPath := NIL
      lChoice   := .f.
   endif
endif

if lChoice
   cLastPath := cPath
   waiton( { "Press desired drive letter", "or press ESC to exit"}, .f.)
   cDrive := upper(chr(inkey(0)))
   waitoff()
   if isalpha(cDrive)
      if ( nHandle := fcreate(cDrive + ":\NUL") ) == -1
         err_msg( { cDrive + ": is not a valid drive" })
      else
         lRetval := .t.
         fclose(nHandle)

         // see if this drive letter was previously stored in aCurDir
         // if so, change its entry to reflect current directory
         // if not, add it now
         nEle := ascan(aCurDir, left(cPath, 1))
         if nEle > 0
            aCurDir[nEle] := cPath
         else
            aadd(aCurDir, cPath)
         endif

         // see if this drive letter was previously stored in aCurDir
         // if so, change path to the last directory visited on that drive
         if ( nEle := ascan(aCurDir, upper(cDrive)) ) > 0
            cPath := aCurDir[nEle]
         else
            cPath := upper(cDrive) + ":\"
         endif

         // if we are inside the APICK() window, force an exit
         if lMustChange == NIL
            keyboard chr(K_ESC)
         endif
      endif
   endif
endif
return lRetval

* end static function ChangeDrive()
*--------------------------------------------------------------------*


/*
   Function:       SHOWKEY()
   Author:         Greg Lief
   Copyright:      (c) 1993 Greg Lief
   Dialect:        Clipper 5.01
   Purpose:        Display key expression of highlighted NTX file
   Syntax:         SHOWKEY( <cFile> )
   Parameters:     <cFile> - path/name of index file
   Returns:        Nothing
   Support Funcs:  GrabKey() - open index file and retrieve key
*/
static function showkey(cFile)
local cKey
local nTop := row() + 1
local nLeft
if ! ("<" $ cFile) .and. ! (".." $ cFile)
   gfsaveenv({ nTop, 10, nTop + 3, maxcol() - 9 }, , 'w/b')
   cKey := GrabKey(cFile)
   if ! empty(cKey)
      // allow enough room for box title
      nLeft := int((maxcol() - max(len(cKey), 14) - 2) / 2)
      shadowbox(nTop, nLeft, nTop + 2, maxcol() - nLeft, 2, "Index Key")
      scrncenter(nTop + 1, cKey)
      ginkey(0)
   endif
   gfrestenv()
endif
return nil

* end static function ShowKey()
*--------------------------------------------------------------------*


/*
   Function:       GRABKEY()
   Author:         Greg Lief
   Copyright:      (c) 1993 Greg Lief
   Dialect:        Clipper 5.x
   Purpose:        Open NTX file with low-level functions to grab index key
   Syntax:         GRABKEY( <cFile> )
   Parameters:     <cFile> - path/name of index file
   Returns:        The index key
   Notes:          Called from SHOWKEY()
*/
static function grabkey(cFile)
local cRetval := space(254)
local nHandle := fopen(cFile, FO_READ)
local cExt    := substr(cFile, at('.', cFile) + 1)
if ferror() == 0
   do case
      case cExt == "NTX"
         fseek(nHandle, 22)      // skip past .NTX header
         fread(nHandle, @cRetval, 254)
      case cExt == "IDX"
         fseek(nHandle, 512)      // skip past .IDX header
         fread(nHandle, @cRetval, 254)
      case cExt == "NDX"
         fseek(nHandle, 24)      // skip past .NDX header
         fread(nHandle, @cRetval, 254)
      //
      // no can do MDX and CDX files
      //
   endcase
   fclose(nHandle)
   cRetval := strtran(cRetval, chr(0), '')
endif
return trim(cRetval)

* end static function GrabKey()
*--------------------------------------------------------------------*

/*
   Function:      SHOWSTRU()
   Filename:      SHOWSTRU.PRG
   Author:        Greg Lief
   Copyright:     (c) 1993 Greg Lief
   Dialect:       Clipper 5.x
   Purpose:       Display structure of highlighted DBF file
   Syntax:        SHOWSTRU( <cFile> [, <cTitle> ] )
   Parameters:    <cFile> is the path/filename of the DBF in question.
   Returns:       Nothing
*/
static function ShowStru(cFile)
local aStru
local oBrowse
local oColumn
local nKey
local nEle
local nMaxEle
local nOldArea
local cAlias := StripPath(cFile)
local hotkey := lastkey()
local bOldblock := setkey(hotkey, NIL)
local lContinue
local lHad2Open := .f.

#ifdef MOUSE
local nMRow
local nMCol
HideMouseCursor()
#endif

if ! ("<" $ cFile) .and. ! (".." $ cFile)
   gfsaveenv( .t., , ColorSet(C_APICK_UNSELECTED, .t.))
   lHad2Open := (select( cAlias ) == 0)
   if lHad2Open
      nOldArea := select()
      lContinue := net_use(cFile, .f.)
   else
      lContinue := .t.
   endif
   if lContinue
      aStru := (cAlias)->( dbstruct() )
      if lHad2Open
         (cAlias)->( dbCloseArea() )
         select(nOldArea)
      endif
      oBrowse := TBrowseNew(1, 17, min(2 + len(aStru), maxrow() - 1), maxcol() - 17)
      shadowbox(0, 16, min(3 + len(aStru), maxrow()), maxcol() - 16, ;
                2, "Structure of " + cAlias)
      nEle := 1    // this will serve as our placeholder in the array
      nMaxEle := len(aStru)
      oBrowse:colSep     := chr(32) + chr(179) + chr(32)
      oBrowse:headSep    := chr(205) + chr(209) + chr(205)
      oBrowse:colorSpec  := ColorSet(C_APICK_UNSELECTED, .t.) + ',' + ;
                            ColorSet(C_APICK_TAGGED, .t.)      + ',' + ;
                            ColorSet(C_MESSAGE, .t.)
      oBrowse:skipBlock  := { | SkipCnt | Gilligan(@nEle, SkipCnt, nMaxEle) }
      oBrowse:goTopBlock := { || nEle := 1 }
      oBrowse:goBottomBlock := { || nEle := nMaxEle }
      oColumn := TBColumnNew("Name",     { || aStru[nEle, DBS_NAME] })
      oColumn:width := 10
      oBrowse:AddColumn(oColumn)
      oColumn := TBColumnNew("Type",     { || aStru[nEle, DBS_TYPE] })
      oBrowse:AddColumn(oColumn)
      oColumn := TBColumnNew("Width",    { || aStru[nEle, DBS_LEN] })
      oBrowse:AddColumn(oColumn)
      oColumn := TBColumnNew("Decimals", { || aStru[nEle, DBS_DEC] })
      oBrowse:AddColumn(oColumn)
      oBrowse:autoLite := .f.
      do while nKey <> K_ESC
         do while ( nKey := inkey() ) == 0 .and. ! oBrowse:stabilize()
         enddo
         if nKey == 0
            // highlight current row
            #ifdef MOUSE
            HideMouseCursor()
            #endif
            oBrowse:colorRect( {oBrowse:rowPos, 1, oBrowse:rowPos, 4}, { 3, 3 })
            oBrowse:hiLite()
            #ifdef MOUSE
            ShowMouseCursor()
            #endif
            nKey := ginkey(0, "KEY")
            oBrowse:deHilite()
         endif

         #ifdef MOUSE
         if nKey == K_LEFT_DBLCLICK
            if GFLastRow() == nMRow .and. GFLastCol() == nMCol
               nKey := K_ENTER
            else
               nKey := K_LEFT_DOWN
            endif
         endif

         do case
            case nKey == K_RIGHT_DOWN
               nKey := K_ESC

            case nKey == K_LEFT_DOWN
               nMRow := GFLastRow()
               nMCol := GFLastCol()

               if nMRow >= oBrowse:nTop .and. nMRow <= oBrowse:nBottom
                  oBrowse:rowPos := nMRow - oBrowse:nTop - 1
                  oBrowse:refreshAll()
               endif
         endcase
         #endif

         do case

            case nKey == K_UP
               oBrowse:refreshCurrent()         // dehilight current row
               oBrowse:up()

            case nKey == K_DOWN
               oBrowse:refreshCurrent()         // dehilight current row
               oBrowse:down()

            case nKey == K_PGUP
               oBrowse:refreshCurrent()         // dehilight current row
               oBrowse:pageUp()

            case nKey == K_PGDN
               oBrowse:refreshCurrent()         // dehilight current row
               oBrowse:pageDown()

            case nKey == K_HOME .or. nKey == K_CTRL_PGUP
               oBrowse:refreshCurrent()         // dehilight current row
               oBrowse:goTop()

            case nKey == K_END .or. nKey == K_CTRL_PGDN
               oBrowse:refreshCurrent()         // dehilight current row
               oBrowse:goBottom()

         endcase
      enddo
   else
      err_msg( { "Could not open " + cFile } )
   endif
   gfrestenv()
endif

#ifdef MOUSE
ShowMouseCursor()
#endif

setkey(hotkey, bOldBlock)
return nil

* end static function ShowStru()
*--------------------------------------------------------------------*

* eof: GFPICKFI.PRG
