/*
   Program: GFPRINTQ.PRG
   Functions to enable a print queue for BACKPRINT() background printing
   System: GRUMPFISH LIBRARY
   Author: Greg Lief
   Numerous additional enhancements by Mark Lukianchuk
   Copyright (c) 1993-4 Greg Lief
   CA-Clipper 5.x version
   Compile instructions: clipper gfprintq /n /w
*/

#include "grump.ch"
#include "shadowb.ch"
#include "inkey.ch"

//these constants delineate the structure of the queue array
#define FILENAME    1
#define PRINTPORT   2
#define BUFFSIZE    3
#define DELETE_FILE 4
#define TITLE       5
#define COPIES      6
#define PRIORITY    7
#define TIME        8
#define DATE        9

static aQueue := {}                              // print queue
static aCurrPrint := {}                          // current job
static cCurrentJob

/*
  Function: GFPQAddJob(<cFile> [,<cPort>] [,<nBuffSize>] [,<lDelete>]
                       [,<cTitle>] [,<nCopies>] [,<lPriority>])

  Parameters:    <cFile>     = file to be printed
                 <cPort>     = printer port (default: LPT1)
                 <nBuffSize> = buffer size (default: 128 bytes)
                 <lDelete>   = delete file when finished printing
                 <cTitle>    = what print job is all about (e.g. Cash Report)
                 <nCopies>   = # of copies to print (default: 1)
                 <lPriority> = priority job? (default: .F.)
                               priority jobs get sent to the top of the queue
*/
function GFPQAddJob(cFile, cPort, nBuffSize, lDelete, cTitle,;
                    nCopies, lPriority)
local nEle := 0
local nLen := len(aQueue)
default cPort to "LPT1"
default nBuffSize to 128
default lDelete to .F.
default cTitle to "<No Title>"
default nCopies to 1
default lPriority to .F.

if lPriority
   asize(aQueue, nLen+1)

   for nEle := nLen+1 to 2 step -1
      aQueue[nEle] := aQueue[nEle-1]               // bump them down
   next

   aQueue[1] := { cFile, cPort, nBuffSize, lDelete, LEFT(cTitle,15),;
                  min(nCopies,99), lPriority, time(), date() }
else
   aadd(aQueue, { cFile, cPort, nBuffSize, lDelete, LEFT(cTitle,15),;
                  min(nCopies,99), lPriority, time(), date() } )
endif

return nil

//-------------------------------------------------------------------

/*
  Function:    GFPQ()
  Parameters:  None
  Notes:       Should be configured to be called automatically by
               GFONTICK(), such as:

               gfontick("gfpq",90) // check print queue every 5 seconds
*/
function GFPQ()
if ! IsPrinting()                                 // not printing anything
   if cCurrentJob <> NIL                          // was a job just printed?
      if aCurrPrint[DELETE_FILE]
         ferase(cCurrentJob)                      // delete if completed
      endif
      if aCurrPrint[COPIES] == 0                  // clear last job when done
         aCurrPrint := {}
         cCurrentJob := NIL
      else                                        // still more copies to go
         aCurrPrint[COPIES]--
         BackPrint(aCurrPrint[FILENAME], aCurrPrint[PRINTPORT],;
                   aCurrPrint[BUFFSIZE])
      endif
   elseif ! empty(aQueue)                         // print another from queue
      aCurrPrint := aQueue[1]
      akill(aQueue,1)                             // move from queue to current

      aCurrPrint[COPIES]--                        // print first copy

      BackPrint(aCurrPrint[FILENAME], aCurrPrint[PRINTPORT],;
                aCurrPrint[BUFFSIZE])

      cCurrentJob := aCurrPrint[FILENAME]
   endif
endif
return nil

//-------------------------------------------------------------------


/*
  Function:    GFPQStatus()
  Parameters:  None
  Notes:       Pop up display to check printing status
*/
function GFPQStatus(nRefresh)
local cBuff
local oBrow := TBrowseNew(9, 3, 19, 77)
local oColumn

local nKey
local nOldCurs
local nEle := 1
local nCount
local cColorSpec  := ColorSet(C_GRUMPBROW_SAY, .T.) + ',' + ;
                     ColorSet(C_GRUMPBROW_GET, .T.) + ',,,'+ ;
                     ColorSet(C_GRUMPBROW_SAY, .T.)
local cBoxColor   := ColorSet(C_GRUMPBROW_BOX, .T.)

default nRefresh to 5

shadowbox buffer cBuff top 5 bottom 20 left 2 right 78 ;
          title "Print Queue Status";
          footer "<DEL> to delete highlighted entry, +/- to change # of copies";
          color cBoxColor

oBrow:headSep := chr(205)+chr(209)+chr(205)
oBrow:colSep := chr(32)+chr(179)+chr(32)
oBrow:colorSpec := cColorSpec

@ 7,  7 SAY "Currently Printing:" COLOR cBoxColor
@ 7, 49 SAY "Copies Remaining: "  COLOR cBoxColor

oBrow:goTopBlock := { | | nEle := 1 }
oBrow:goBottomBlock := { | | nEle := len(aQueue) }
oBrow:skipBlock := { | nSkip | Gilligan(@nEle, nSkip, len(aQueue)) }

oColumn := TBColumnNew("TITLE", {|| aQueue[nEle, TITLE] })
oColumn:width := 15
oBrow:AddColumn(oColumn)

oColumn := TBColumnNew("FILE", {|| aQueue[nEle, FILENAME] })
oColumn:width := 12
oBrow:AddColumn(oColumn)

oColumn := TBColumnNew("DATE", {|| aQueue[nEle, DATE] })
oBrow:AddColumn(oColumn)

oColumn := TBColumnNew("TIME", {|| aQueue[nEle, TIME] })
oBrow:AddColumn(oColumn)

oColumn := TBColumnNew("DEL", {|| IIF(aQueue[nEle, DELETE_FILE],"  ","   ") })
oBrow:AddColumn(oColumn)

#ifdef CLIPPER52
   oColumn := TBColumnNew("COPY", {|| aQueue[nEle, COPIES] })
   oColumn:picture := "9999"
#else
   oColumn := TBColumnNew("COPY", {|| TRANSFORM(aQueue[nEle, COPIES], '9999')})
#endif

oColumn:width := 4
oBrow:AddColumn(oColumn)

oColumn := TBColumnNew("PRI", {|| IIF(aQueue[nEle, PRIORITY],"  ","   ") })
oBrow:AddColumn(oColumn)

nOldCurs := setcursor()
setcursor(0)

nCount := seconds()

while nKey <> K_ESC

   if len(aQueue)==0
      nKey := inkey()
   else
      while ((nKey := inkey())==0) .and. ! oBrow:stabilize()
      end
   endif

   @ 7, 27 SAY PADR( IIF(cCurrentJob == NIL, "<NONE>", ;
                     IIF(aCurrPrint[TITLE] == "<No Title>", cCurrentJob, ;
                     aCurrPrint[TITLE])), 15) COLOR cBoxColor

   @ 7, 67 SAY IIF(cCurrentJob==NIL,"N/A",;
           PAD(LTRIM(STR(aCurrPrint[COPIES]+1)),3)) COLOR cBoxColor

   if len(aQueue) == 0

      @ 11, 3 SAY replicate("",29) + "Print Queue Empty" + ;
                  replicate("",29) COLOR cBoxColor

   else

      // auto-refresh logic
      if seconds() - nCount > nRefresh
         oBrow:refreshAll()
         #ifdef CLIPPER52
         oBrow:forceStable()
         #else
         while ! oBrow:stabilize()
         end
         #endif
         nCount := seconds()
      endif

      do case

         case nKey == K_UP
            oBrow:up()

         case nKey == K_DOWN
            oBrow:down()

         case nKey == ASC("-")
            if nEle > 0
               aQueue[nEle][COPIES] := MAX(1, aQueue[nEle][COPIES]-1)
               oBrow:refreshAll()
            endif

         case nKey == ASC("+")
            if nEle > 0
               aQueue[nEle][COPIES] := MIN(99, aQueue[nEle][COPIES]+1)
               oBrow:refreshAll()
            endif

         case nKey == K_DEL
            if (nEle > 0) .and. yes_no("Delete this job")
               akill(aQueue, nEle)
               oBrow:refreshAll()
            endif

      endcase

   endif

end
setcursor(nOldCurs)
byebyebox(cBuff)
return nil

// end of file GFPRINTQ.PRG
