/*
   Program: HELP.PRG
   System: GRUMPFISH LIBRARY
   Author: Greg Lief
   Copyright (c) 1988-93, Greg Lief
   CA-Clipper 5.x Version
   Compile instructions: clipper help /n /w

   Displays user-defined help screens

   Procs & Fncts: HELP
                : SHOWHELP()
                : PRINTHELP()
                : HELPHEAD()
*/

#define HELP_CODE 5

#include "apick.ch"
#include "grump.ch"
#include "inkey.ch"
#include "achoice.ch"

static cMainScreen                 // saved entry screen
static helpfile := "help"     // default help file

function help(cProc, nLine, cVar)
local nWorkArea := select()
local bOldF1
local nHotKey
local bOldBlock
local oGet
memvar helpcode        // would be declared in calling program if it exists
GFSaveEnv( , 0)        // shut off cursor

#ifdef MOUSE
HideMouseCursor()
#endif

cMainScreen := savescreen()

#ifdef MOUSE
ShowMouseCursor()
#endif

/*
   usually if HELPDEV() was called from anything other than GINKEY(),
   cProc will be a null string.  We would then have to use the procedure
   name three levels deep.  However, in the event that the procedure
   name passed to HELPDEV() is an internal Clipper function (beginning
   with double underscore "__"), then we must use the procedure name
   four levels deep.
*/
if empty(cProc)
   cProc := procname(3)
elseif left(cProc, 2) = '__'
   cProc := procname(4)
endif
bOldBlock := setkey(nHotKey := lastkey(), NIL)

// set F1 key to show help index
bOldF1 := setkey(K_F1, {|p, l, v| showindex(p, l, v)} )

// make sure everything is proper before proceeding
if ! file(helpfile + '.dbf')
   err_msg( { 'Help files missing' } )
elseif net_use(helpfile + " ALIAS HELPME", .f.)
   // rebuild index if necessary
   if ! file(helpfile + indexext())
      if flock()
         waiton("Indexing help database")
         index on _field->theproc + _field->var to (helpfile)
         unlock
      else
         waiton({ "Help database is being reindexed", ;
                  "Please wait or press Esc to abort" } )
         #ifdef MOUSE
         do while ! flock() .and. inkey() <> K_ESC .and. lastkey() <> K_RIGHT_DOWN
         #else
         do while ! flock() .and. inkey() <> K_ESC
         #endif
         end
      endif
      waitoff()
   else
      set index to (helpfile)
   endif

   // check to see if there is an active GET, and if so, check
   // whether there is a help code associated with it
   if ( oGet := getactive() ) <> NIL .and. valtype(oGet:cargo) == "A" .and. ;
              len(oGet:cargo) == HELP_CODE .and. oGet:cargo[HELP_CODE] <> NIL
      cVar := upper(oGet:cargo[HELP_CODE])
   endif

   // seek on helpcode first if such a variable is defined, first
   // ensuring that there is an active index (w/ INDEXKEY())!
   if ! empty(indexkey(0))
      if dbseek( if(type('helpcode') <> "U", helpcode, padr(cProc, 10) + cVar) )
         showhelp()
      else
         showindex()
      endif
   endif
   dbclosearea()
endif
GFRestEnv()
setkey(K_F1, bOldF1)          // restore F1 hot-key
setkey(nHotKey, bOldBlock)    // restore hot-key
restscreen(,,,, cMainScreen)
select(nWorkArea)
return NIL

* end of function Help() (i need somebody...)
*--------------------------------------------------------------------


/*
   Function: ShowHelp()  --> display help screen
*/
static function ShowHelp
local nMid
shadowbox(helpme->toprow, helpme->lt_col, helpme->botrow, ;
     helpme->rt_col, helpme->boxno,,, color_n2s(helpme->boxcolor))
nMid := helpme->lt_col + (int(helpme->rt_col - helpme->lt_col) / 2)
@ helpme->botrow, nMid - int(len(trim(helpme->footer)) / 2) ;
       ssay trim(helpme->footer) color color_n2s(helpme->ftcolor)
@ helpme->toprow, nMid - int(len(trim(helpme->title)) / 2) ;
       ssay trim(helpme->title) color color_n2s(helpme->titcolor)
setcolor(color_n2s(helpme->txtcolor))
memoedit(helpme->text, helpme->toprow + 1, helpme->lt_col + 1, ;
         helpme->botrow - 1, helpme->rt_col - 1, .f., 'PrintHelp')
return NIL

* end of static function ShowHelp() (not just anybody...)
*--------------------------------------------------------------------


/*
   Function: ShowIndex()  --> display help screen index
*/
static function ShowIndex(cProc, nLine, cVar)
static aHelpIndex := {}    // array to hold help screen titles
static aHelpRecs           // array to hold help screen record numbers
static lIndexentry
local nMarker := recno()
local cBuffer
local oldcolor
local ele := 1
local nHotKey

// if we got here from a hot-key, restore the original screen first!
if cProc <> NIL
   setkey(nHotKey := lastkey(), NIL)
   cBuffer := savescreen()
   restscreen(,,,, cMainScreen)    // original screen
endif
go top

// load lIndexentry flag and help index arrays on first pass only...
// NOTE: lIndexentry is for downward compatibility only because
// older versions of HELP.DBF will not have the INDEXENTRY field
if lIndexentry == NIL
   lIndexentry := ( helpme->( fieldpos("indexentry") ) > 0 )
   waiton( { "Creating help index" } )
   do while ! eof()
      if ! lIndexentry .or. _field->indexentry
         aadd( aHelpIndex, { _field->title, recno() } )
      endif
      skip
   enddo
   if ! empty(aHelpIndex)
      // sort this based on the title
      asort(aHelpIndex,,, { | a, b | a[1] < b[1] } )
      aHelpRecs := array(len(aHelpIndex))

      // make this array single-dimension and create parallel
      // array of record numbers
      aeval(aHelpIndex, { | ele, num | aHelpRecs[num] := ele[2], ;
                                       aHelpIndex[num] := ele[1] } )
   endif
   waitoff()
endif
if empty(aHelpIndex)
   err_msg({ "No help index available" })
else
   do while ele > 0
      apick ele array aHelpIndex initial ele title "Help Index"
      if ele > 0
         go aHelpRecs[ele]
         restscreen(,,,, cMainScreen)    // original screen
         ShowHelp()
         restscreen(,,,, cMainScreen)    // original screen
      endif
   enddo
endif
go nMarker

// restore screen and hot key if we got here from a hot key
if cProc <> NIL
   restscreen(,,,, cBuffer)
   setkey(nHotKey, {|p, l, v| showindex(p, l, v)} )
endif
setcolor(oldcolor)
return NIL

* end of static function ShowIndex() (ya know I need someone...)
*--------------------------------------------------------------------*


/*
  function: PrintHelp() -- UDF for MEMOEDIT() above
*/
function PrintHelp(mstat, nLine, mcol)
local key := lastkey()
local linewidth
local lines
local currline
local ret_val := 0
if key == K_ALT_P    // print this help screen
   if PrintOK()
      #ifdef MOUSE
      waiton('Now printing help text, press Esc or right-click to abort')
      #else
      waiton('Now printing help text, press Esc to abort')
      #endif
      set device to printer
      currline := 1
      linewidth := helpme->rt_col - helpme->lt_col + 1
      lines := mlcount(helpme->text, linewidth)
      HelpHead(.t.)
      #ifdef MOUSE
      while currline <= lines .and. inkey() <> K_ESC .and. lastkey() <> K_RIGHT_DOWN
      #else
      while currline <= lines .and. inkey() <> K_ESC
      #endif
         @ prow()+1,helpme->lt_col say ;
                     trim(memoline(helpme->text, linewidth, currline++))
         if prow() >= 59
            HelpHead(.f.)
         endif
      end
      eject
      set device to screen
      waitoff()
   endif

#ifdef MOUSE
elseif key == K_RIGHT_DOWN
   ret_val := K_ESC
#endif

endif
return ret_val

* end function PrintHelp()
*--------------------------------------------------------------------*


/*
  function: HelpHead() -- heading when printing help text
*/
static function HelpHead(lReset)
static nPage := 1
if lReset
   nPage := 1
endif
@ 0,0 say ''
@ 1,1 say upper(helpme->title)
@ 1,72 say 'Page ' + ltrim(str(nPage++))
@ 3, 0 say ''
return NIL

* end static function HelpHead()
*--------------------------------------------------------------------*


/*
  Function: HelpSet() -- change name of help database file (default HELP.DBF)
  Syntax:   HelpSet(<cNewfile>)
  Returns:  Previous help file name
*/
function HelpSet(cNewFile)
local cOldFile := helpfile
helpfile := cNewFile
return cOldFile

* end function HelpSet()
*--------------------------------------------------------------------*

* end of file HELP.PRG
