/*
   Program: INDEXBAR()
   System: GRUMPFISH LIBRARY
   Author: Greg Lief
   Copyright (c) 1988-93, Greg Lief
   CA-Clipper 5.x version
   Compile instructions: clipper indexbar /n /w

   Procs & Fncts: IndexBar()
                : GFShowBar()

           Calls: SHADOWBOX()   (function in SHADOWB.PRG)

   IndexBar() displays a graph showing percentage complete when
   indexing files.  When indexing is complete, IndexBar() will
   then strip the graph UDF() call from the index file so that you
   don't have to look at the graph when the index is updated.

   IMPORTANT: If you wish to make use of specialized CA-Clipper 5.2
   indexing (e.g., order bags, conditional, etcetera), you should
   recompile this file with the /dCLIPPER52 command-line parameter!
*/

#include "grump.ch"
#include "fileio.ch"

static nRecs        // total number of records in current database
                    // I put it here instead of referring to RECCOUNT()
                    // because it speeds things up a klick or two

static nGraphLen    // must be re-initialized each time INDEXBAR() is called

/*
    Function: IndexBar()
    Syntax:   IndexBar(<cFile>, <cKey>, <nRow>, <lShowcount>, <lUnique>, <cMsg>)
              <cFile> = name of index file... to be displayed FYI
              <cKey>  = self-explanatory -- if not passed, will
                        be assumed to be the same as the index filename
              <nRow> = row at which to place window (default: centered)
              <lShowcount> = whether or not to display record counter
              <lUnique> = whether or not to create UNIQUE index
              <cMsg>    = message to use rather than the default
                          (which is "now creating index file <cfile>")
    Example:  IndexBar('customer', 'lname')
    Returns:  Logical: .T. if index is built, .F. if not
    Notes:    will not process empty database files
*/
function indexbar(cFile, mkey, nRow, showcount, lUnique, cMsg, ;
                  cFor, bFor, lAll, bWhile, nNext, nRec, nEvery, ;
                  lRest, lDescend, cOrderName)
local lRetVal := .f.
local xx
local oldscrn
local bkey
local nLeft
default mkey to cFile
default showcount to .f.
default lUnique to .f.
nGraphLen := 0
GFSaveEnv(, 0)
// make sure that the index key expression is valid to preclude a crash!
if ! valtype(mkey) $ "UE"
   bKey := makeblock(mkey)       // see GRUMP.CH for MakeBlock()
   default nRow to ( maxrow() / 2 ) - 3
   dbClearIndex()
   // no point in going thru this rigmarole on an empty database (which
   // would give us zero divide, a/k/a "the great divide")
   if lastrec() > 0
      nRecs := reccount()
      // if no extension specified, add default now
      if ! "." $ cFile
         AddExtension(cFile, substr(indexext(), 2))
      endif
      ColorSet(C_MESSAGE)
      nLeft := ( maxcol() / 2 ) - 30
      dispbegin()
      oldscrn := shadowbox(nRow, nLeft, nRow + 6, nLeft + 63, 2)
      @ nRow + 3, nLeft + 2  ssay ''
      @ nRow + 3, nLeft + 61 ssay ''
      @ nRow + 4, nLeft + 2  ssay '' + replicate("", 9) + ';'
      @ nRow + 5, nLeft + 2  ssay '0'
      for xx := 1 to 10
         @ nRow + 5, nLeft + xx * 6 ssay str(10 * xx, 3)
      next
      SCRNCENTER(nRow + 1, if(cMsg <> NIL, cMsg, "Creating index file " + ;
                              cFile + if(cOrderName <> NIL, " (" + ;
                                      cOrderName + " tag)", "") + "...") )
      ColorSet(C_APICK_STATUSBAR)
      @ nRow + 3, nLeft + 3 ssay replicate(chr(177), 58)
      setpos(nRow + 3, nLeft + 3)
      dispend()
      ColorSet(C_APICK_INDICATOR)
      if bWhile == NIL
         dbGoTop()
      endif

      #ifdef CLIPPER52

      // if anything beyond the sixth parameter was received, that
      // means that we are using RDD indexing (e.g., conditional,
      // order bags, etcetera) and must therefore use the 5.2
      // functions instead of good ole dbCreateIndex()
      if pcount() > 6
         ordCondSet(cFor, bFor, lAll, bWhile, ;
                    { || gfshowbar(NIL, recno(), showcount) }, ;
                    nEvery, recno(), nNext, nRec, lRest, lDescend)
         ordCreate(cFile, cOrderName, mkey, bkey, lUnique)
         ordListClear()
         ordListAdd(cFile)
      else

      #endif

         /*
            I know it looks dumb, but the parentheses around the indexkey
            (2nd parameter) force dbCreateIndex() to evaluate the block
            (3rd parameter) for each record in the DBF.  If you remove
            the parentheses, the block will only be evaluated once, which
            means that the status bar will not be drawn.
         */
         dbCreateIndex(cFile, "(" + mkey + ")", ;
                       {|| gfshowbar(bkey, recno(), showcount)}, lUnique)
         dbClearIndex()
         dbSetIndex(cFile)

      #ifdef CLIPPER52

      endif

      #endif

      byebyebox(oldscrn)
   else
      #ifdef CLIPPER52

      if pcount() > 6
         ordCondSet(cFor, bFor, lAll, bWhile, , , recno(), ;
                    nNext, nRec, lRest, lDescend)
         ordCreate(cFile, cOrderName, mkey, bkey, lUnique)
         ordListClear()
         ordListAdd(cFile)
      else

      #endif

      dbCreateIndex(cFile, mkey, bkey, lUnique)
      dbClearIndex()
      dbSetIndex(cFile)

      #ifdef CLIPPER52

      endif

      #endif

   endif
   lRetVal := .t.
endif
GFRestEnv()
return lRetVal

* end function IndexBar()
*--------------------------------------------------------------------*


/*
    Function: GFShowBar()
    Syntax:   GFShowBar(<index key>)
              <index key> = self-explanatory
    Purpose:  Display status graph whilst indexing
    Example:  See above -- NOT MEANT FOR USE AS A STANDALONE
    Returns:  the index key for each record
*/
static function gfshowbar(bkey, nCurrentRec, showcount)
local oldcol
if nCurrentRec <= nRecs
   if showcount
      ColorSet(C_MESSAGE)
      oldcol := col()  // save column position for restoration below
      SCRNCENTER(row() - 1, if(nCurrentRec <> nRecs, ;
                               "Indexing Record " + ltrim(str(nCurrentRec)) + ;
                               " of " + ltrim(str(nRecs)), ;
                               space(18) + "Finished !" + space(18)))
      setpos(row()+1, oldcol)     // reset to proper column
      ColorSet(C_APICK_INDICATOR)
   endif

   // draw more of the graph if the relative position has changed
   do while nGraphLen < int(nCurrentRec / nRecs * 58)
      dispout(chr(219))
      nGraphLen++
   enddo
endif
#ifdef CLIPPER52
   return if(bKey == NIL, .t., eval(bKey))
#else
   return eval(bKey)
#endif

* end static function GFShowBar()
*--------------------------------------------------------------------*

* eof indexbar.prg
