/*
    Program: MENUV()
    System: GRUMPFISH LIBRARY
    Author: Greg Lief
    Copyright (c) 1988-93, Greg Lief
    CA-Clipper 5.x version
    Compile instructions: clipper menuv /n/w/a
    Compile with "/dMOUSE" for mouseability

    Painless vertical bounce-bar menu

    Sample format of array to pass to this function:

    mainmenu := { { 'Data Entry', 'Edit info',    { || data_entry() } }, ;
                  { 'Reports', 'Hard copies',     { || reports() } }   , ;
                  { 'Utilities', 'Miscellaneous', { || utilities() } } , ;
                  { 'Quit', 'Exit to DOS'} }

    To skip anything, just leave it NIL.  For example:

                  { 'Reports', , }  , ;

    will display Reports as a menu option, but will not display
    an accompaying message nor execute any function upon selection.
*/

#include "grump.ch"

MEMVAR promptList

#ifdef MOUSE
#include "inkey.ch"

// structure of aCoords nested arrays
#define A_TOP   1
#define A_LEFT  2
#define A_RIGHT 3

#endif

// structure of aArray nested arrays
#define M_PROMPT   1
#define M_MESSAGE  2
#define M_ACTION   3


function Menuv(aArray, cTitle, nBoxType, cBoxColor, cTitleColor, nTop, nLeft)

local nChoice
local lUseProc
local nElements := len(aArray)
local nMaxLen
local nBottom
local nRight
local gfmenuscrn
local xx
local cProc
local lOldMsgCtr
local nOldMsgRow
local lOldWrap
local nMaxRow := maxrow() + 1
local nMaxCol := maxcol() + 1
local lMainLoop := .t.
#ifdef MOUSE
   local aCoords
   local lSubLoop
   local lFirstLoop := .t.
   local cSelected
   local cUnselected
   local nKey
   local nOldChoice
   local nMsgRow
   local nMaxMsgWidth := 0
   local nMsgLeft
   local cSpacing
#endif
// we will only center the prompts if the left coord was not passed
local lCenterPrompts := (nLeft == NIL)

#ifdef MOUSE
   GFSaveEnv(, 0)  // must shut off cursor if not using stock MENU TO!
#else
   GFSaveEnv()
#endif

// establish MESSAGE and WRAP settings
lOldMsgCtr := set(_SET_MCENTER, .T.)  // set message to be centered

// if no message row has been established already, set it to 24
if (nOldMsgRow := set(_SET_MESSAGE)) == 0
   set(_SET_MESSAGE, nMaxRow)
endif
lOldWrap := set(_SET_WRAP, .T.)     // SET WRAP ON

// establish defaults if parameters were not passed
default cTitle to 'Menu'
default nBoxType to 1
default cBoxColor to ColorSet(C_MENU_UNSELECTED, .T.) + ',' + ;
                     ColorSet(C_MENU_SELECTED, .T.)
default cTitleColor to ColorSet(C_MENU_SELECTED, .T.)

// limit # of menu items to 4 less than total number of rows
if nElements < nMaxRow - 4
   // determine maximum length for menu selections and draw box accordingly
   nMaxLen := len(cTitle) + 4  // must be at least as wide as the menu title!
   lUseProc := .f.             // set true if user passed procedures to do
   aeval(aArray, { | a | nMaxLen := max(nMaxLen, len(a[M_PROMPT])) } )
   // if any procedure names were passed, set lUseProc true
   lUseProc := ( ascan( aArray, { | a | a[M_ACTION] <> NIL } ) > 0)

   // assign left and right column coordinates
   if nLeft == NIL
      nLeft := int((nMaxCol - nMaxLen) / 2)
   else
      nLeft := min(nLeft, nMaxCol - nMaxLen - 1)
   endif
   nRight := nLeft + nMaxLen

   // calculate top and bottom rows for box based on # of options
   if nTop == NIL
      nTop := int(nMaxRow/2) - int((nElements + 3) / 2)
   else
      nTop := min(nTop, nMaxRow - nElements - 3)
   endif
   nBottom := nTop + nElements + 2

   if nBoxType > 5
      shadowbox(nTop, nLeft-1, nBottom, nRight, (nBoxType-1) % 5 + 1, ;
                cTitle, , cBoxColor)
   else
      @ nTop, nLeft-1, nBottom, nRight box BOXFRAMES[nBoxType] color cBoxColor
      @ nTop, nLeft + int((nRight - nLeft - len(cTitle)) / 2) ;
              ssay ' ' + cTitle + ' ' color cTitleColor
   endif

   do while lMainLoop
      #ifdef MOUSE

      if lFirstLoop
         setcolor(cBoxColor)
         // display all prompts centered on the screen
         setpos(nTop + 1, nLeft)
         aCoords := array(nElements)
         nMsgRow := set(_SET_MESSAGE)
         cSelected   := ColorSet(C_MENU_SELECTED, .t.)
         cUnselected := ColorSet(C_MENU_UNSELECTED, .t.)
         for xx := 1 to nElements
            @ row()+1, if(! lCenterPrompts, nLeft, ;
                          int((nMaxCol - len(aArray[xx][M_PROMPT])) / 2)) ;
                       say aArray[xx][M_PROMPT] color cUnselected
            if aArray[xx][M_MESSAGE] <> NIL
               nMaxMsgWidth := max(nMaxMsgWidth, len(aArray[xx][M_MESSAGE]))
            endif
            aCoords[xx] := { row(), col() - len(aArray[xx][M_PROMPT]), col()-1}
         next
         nChoice := 1
         lFirstLoop := .f.
         nMsgLeft := int( ( (nMaxCol + 1) - nMaxMsgWidth ) / 2 )
      endif
      lSubLoop := .t.
      do while lSubLoop
         @ aCoords[nChoice][A_TOP], aCoords[nChoice][A_LEFT] ;
                           ssay aArray[nChoice][M_PROMPT] color cSelected
         if aArray[nChoice][M_MESSAGE] <> NIL
            cSpacing := space( ( nMaxMsgWidth - ;
                                 len(aArray[nChoice][M_MESSAGE]) ) / 2)
            @ nMsgRow, nMsgLeft ssay cSpacing + aArray[nChoice][M_MESSAGE] + ;
                                    cSpacing color cUnselected
         endif

         nOldChoice := 0

         nKey := ginkey( 0 )

         do case
            case nKey == K_UP
               nOldChoice := nChoice

               if nChoice == 1
                  nChoice := nElements
               else
                  nChoice--
               endif

            case nKey == K_DOWN
               nOldChoice := nChoice

               if nChoice == nElements
                  nChoice := 1
               else
                  nChoice++
               endif

            case nKey == K_PGUP .and. nChoice <> 1
               nOldChoice := nChoice ; nChoice := 1

            case nKey == K_PGDN .and. nChoice <> nElements
               nOldChoice := nChoice ; nChoice := nElements

            case nKey == K_ENTER
               lSubLoop := .f.

            case nKey == K_ESC
               lSubLoop := .f.
               nChoice := 0

            case nKey == K_LEFT_DBLCLICK
               lSubLoop := .f.

            case nKey == K_LEFT_DOWN
               xx := 0
               while ++xx <= nElements .and. ;
                             (aCoords[xx][A_TOP] <>  GFLastRow() .or. ;
                              aCoords[xx][A_LEFT] >  GFLastCol() .or. ;
                              aCoords[xx][A_RIGHT] < GFLastCol() )
               end

               if xx <= nElements
                  nOldChoice := nChoice
                  nChoice := xx
               endif

            case nKey == K_RIGHT_DOWN
               nChoice := 0
               lSubLoop := .f.

         endcase

         // redisplay previous prompt in unselected color
         if nOldChoice <> 0
            @ aCoords[nOldChoice][A_TOP], aCoords[nOldChoice][A_LEFT] ;
                      say aArray[nOldChoice][M_PROMPT] color cUnselected
         endif
      enddo

      #else

      setcolor(cBoxColor)
      // first -- display all prompts centered on the screen
      setpos(nTop + 1, nLeft)
      for xx := 1 to nElements
         @ row()+1, if(! lCenterPrompts, nLeft, ;
                       int((nMaxCol - len(aArray[xx][M_PROMPT])) / 2)) ;
                    prompt aArray[xx][M_PROMPT] message aArray[xx][M_MESSAGE]
      next
      menu to nChoice

      #endif

      // exit the loop if they selected the last choice, escaped, or
      // no procedures were passed to be done
      if nChoice == len(aArray) .or. nChoice == 0 .or. ! lUseProc
         lMainLoop := .f.
      else
         // execute the function tied to this option if there is one
         if aArray[nChoice][M_ACTION] <> NIL
            // convert to code block if not passed as code block
            if valtype(aArray[nChoice][M_ACTION]) <> "B"
               cProc := &( "{ || " + aArray[nChoice][M_ACTION] + "}" )
            else
               cProc := aArray[nChoice][M_ACTION]
            endif
            gfmenuscrn := savescreen()
            eval(cProc)
            restscreen(,,,, gfmenuscrn)
         endif
      endif
   enddo
endif
// restore previous message and wrap settings
set(_SET_MCENTER, lOldMsgCtr)
set(_SET_MESSAGE, nOldMsgRow)
set(_SET_WRAP, lOldWrap)
GFRestEnv()
return nChoice

* end function MenuV()
*--------------------------------------------------------------------*

* eof menuv.prg
